@@ -15,10 +15,12 @@ from dateutil import relativedelta as rdelta
 import logging
 from functools import partial
 
-from xlrd import xldate_as_tuple
 import calendar
+
+from openpyxl import Workbook
 from scipy.interpolate import interp1d
 import json
+from typing import Callable
 
 __author__ = 'schien'
 
@@ -393,7 +395,8 @@ class ConstantUncertaintyExponentialGrowthTimeSeriesGenerator(DistributionFuncti
 
         a = growth_coefficients(start_date, end_date, ref_date, alpha, self.size)
 
-        values *= a.ravel()
+        x = a.ravel()
+        values = np.multiply(values, x)
 
         # df = pd.DataFrame(values)
         # df.columns = [kwargs['name']]
@@ -634,35 +637,12 @@ class TableHandler(object):
         self.version = version
 
     @abstractmethod
-    def load_definitions(self, sheet_name, filename=None):
+    def load_definitions(self, sheet_name, filename=None, id_flag=False):
         raise NotImplementedError()
 
 
-class OpenpyxlTableHandler(TableHandler):
-    def load_definitions(self, sheet_name, filename=None):
-        definitions = []
-
-        from openpyxl import load_workbook
-        wb = load_workbook(filename=filename, data_only=True)
-        _sheet_names = [sheet_name] if sheet_name else wb.sheetnames
-        for _sheet_name in _sheet_names:
-            sheet = wb.get_sheet_by_name(_sheet_name)
-            rows = list(sheet.rows)
-            header = [cell.value for cell in rows[0]]
-
-            if header[0] != 'variable':
-                continue
-
-            for row in rows[1:]:
-                values = {}
-                for key, cell in zip(header, row):
-                    values[key] = cell.value
-                definitions.append(values)
-        return definitions
-
-
 class Xlsx2CsvHandler(TableHandler):
-    def load_definitions(self, sheet_name, filename=None):
+    def load_definitions(self, sheet_name, filename=None, id_flag=False):
         from xlsx2csv import Xlsx2csv
         data = Xlsx2csv(filename, inmemory=True).convert(None, sheetid=0)
 
@@ -698,7 +678,7 @@ class DictReaderStrip(csv.DictReader):
 
 
 class CSVHandler(TableHandler):
-    def load_definitions(self, sheet_name, filename=None):
+    def load_definitions(self, sheet_name, filename=None, id_flag=False):
         reader = DictReaderStrip(open(filename), delimiter=',')
 
         definitions = []
@@ -760,14 +740,14 @@ class PandasCSVHandler(TableHandler):
         except AttributeError:
             return text
 
-    def load_definitions(self, sheet_name, filename=None):
+    def load_definitions(self, sheet_name, filename=None, id_flag=False):
         self.version = 2
 
         import pandas as pd
         df = pd.read_csv(filename, usecols=range(15), index_col=False, parse_dates=['ref date'],
                          dtype={'initial_value_proportional_variation': 'float64'},
                          dayfirst=True,
-                         # date_parser=lambda x: pd.datetime.strptime(x, '%d-%m-%Y')
+                         # date_parser=l0ambda x: pd.datetime.strptime(x, '%d-%m-%Y')
                          )
         df = df.dropna(subset=['variable', 'ref value'])
         df.fillna("", inplace=True)
@@ -775,75 +755,143 @@ class PandasCSVHandler(TableHandler):
         return df.to_dict(orient='records')
 
 
-class XLRDTableHandler(TableHandler):
+class OpenpyxlTableHandler(TableHandler):
     version: int
 
+    def __init__(self, version=2):
+        super().__init__(version=version)
+        self.highest_id = -1
+        self.id_map = defaultdict(lambda: defaultdict(dict))
+        self.id_column = None
+
     @staticmethod
     def get_sheet_range_bounds(filename, sheet_name):
-        import xlrd
-        wb = xlrd.open_workbook(filename)
-        sheet = wb.sheet_by_name(sheet_name)
-        rows = list(sheet.get_rows())
+        import openpyxl
+        wb = openpyxl.load_workbook(filename)
+        sheet = wb[sheet_name]
+        rows = list(sheet.iter_rows())
         return len(rows)
 
-    def find_highest_id(self):
+    def add_ids(self, ws=None, values=None, definitions=None, row_idx=None, id_flag=False,
+                sheet_name=None, **kwargs):
         """
-        stub for id management
+        using the id map, assign ids to those variables that have not got an id yet
         :return:
         :rtype:
         """
-        pass
+        name = values["variable"]
+        scenario = values['scenario'] if values['scenario'] else "default"
+        if name not in self.id_map.keys() or scenario not in self.id_map[name].keys():
+            # If this is the first process and it has no ID, set it to 0
+            pid = self.highest_id + 1  # else set it to the highest existing ID plus 1
+            self.highest_id += 1
+            self.id_map[name][scenario] = pid
+            values["id"] = pid
+            logger.debug(f'{name} {scenario}: {values}')
+            definitions[name][scenario]["id"] = pid
+            c = ws.cell(row=row_idx + 2, column=self.id_column)
+            c.value = pid
+            logger.info("ID " + str(pid) + " given to process " + values['variable'])
 
-    def table_visitor(self, wb, _sheet_names, function):
+    def ref_date_handling(self, values: Dict = None, definitions=None, sheet_name=None, id_flag=None,
+                          **kwargs):
+
+        if 'ref date' in values and values['ref date']:
+            if isinstance(values['ref date'], datetime.datetime):
+                # values['ref date'] = datetime.datetime(*xldate_as_tuple(values['ref date'], wb.datemode))
+                if values['ref date'].day != 1:
+                    logger.warning(f'ref date truncated to first of month for variable {values["variable"]}')
+                    values['ref date'] = values['ref date'].replace(day=1)
+            else:
+                raise Exception(
+                    f"{values['ref date']} for variable {values['variable']} is not a date - "
+                    f"check spreadsheet value is a valid day of a month")
+
+        logger.debug(f'values for {values["variable"]}: {values}')
+        name = values['variable']
+        scenario = values['scenario'] if values['scenario'] else "default"
+        # store id's in a map to identify largest existing id
+        if id_flag:
+            if 'id' in values.keys() and (values["id"] or values["id"] == 0):
+                pid = values['id']
+                if name not in self.id_map.keys() or scenario not in self.id_map[name].keys():
+                    # raises exception if the ID already exists
+                    if (any(pid in d.values() for d in self.id_map.values())):
+                        raise Exception("Duplicate ID variable " + name)
+                    else:
+                        self.id_map[name][scenario] = pid
+                        if pid > self.highest_id:
+                            self.highest_id = pid
+        if scenario in definitions[name].keys():
+            logger.error(
+                f"Duplicate entry for parameter "
+                f"with name <{values['variable']}> and <{scenario}> scenario in sheet {sheet_name}")
+            raise ValueError(
+                f"Duplicate entry for parameter "
+                f"with name <{values['variable']}> and <{scenario}> scenario in sheet {sheet_name}")
+        else:
+            definitions[name][scenario] = values
+
+    def table_visitor(self, wb: Workbook = None, sheet_names: List[str] = None, visitor_function: Callable = None,
+                      definitions=None, id_flag=False):
         """
         stub for id management
 
+        :param definitions:
         :param wb:
         :type wb:
-        :param _sheet_names:
-        :type _sheet_names:
-        :param function:
-        :type function:
+        :param sheet_names:
+        :type sheet_names:
+        :param visitor_function:
+        :type visitor_function:
         :return:
         :rtype:
         """
-
-        for _sheet_name in _sheet_names:
+        if not sheet_names:
+            sheet_names = wb.sheetnames
+        for _sheet_name in sheet_names:
             if _sheet_name == 'metadata':
                 continue
-            sheet = wb.sheet_by_name(_sheet_name)
-            rows = list(sheet.get_rows())
+            sheet = wb[_sheet_name]
+            rows = list(sheet.iter_rows())
             header = [cell.value for cell in rows[0]]
-
             if header[0] != 'variable':
                 continue
+            if id_flag:
+                # get the id column number
+                self.id_column = header.index('id') + 1
 
             for i, row in enumerate(rows[1:]):
                 values = {}
                 for key, cell in zip(header, row):
                     values[key] = cell.value
                 if not values['variable']:
-                    # logger.debug(f'ignoring row {i}: {row}')
+                    logger.debug(f'ignoring row {i}: {row}')
                     continue
 
-                function()
+                visitor_function(ws=sheet, values=values, definitions=definitions,
+                                 row_idx=i, sheet_name=_sheet_name, id_flag=id_flag, row=row,
+                                 header=header)
+        return definitions
 
-    def load_definitions(self, sheet_name, filename=None):
-        import xlrd
-        wb = xlrd.open_workbook(filename)
-        sh = None
-
-        definitions = []
-
-        _definition_tracking = defaultdict(dict)
-
-        _sheet_names = [sheet_name] if sheet_name else [sh.name for sh in wb.sheets()]
+    def load_definitions(self, sheet_name, filename: str = None, id_flag=False):
+        """
+        @todo - document that this not only loads definitions, but also writes the data file, if 'id' flag is True
+        :param sheet_name:
+        :param filename:
+        :param id_flag:
+        :return:
+        """
+        import openpyxl
+        wb = openpyxl.load_workbook(filename, data_only=True)
 
+        definitions = defaultdict(lambda: defaultdict(dict))
+        _sheet_names = [sheet_name] if sheet_name else wb.sheetnames
         version = 1
 
         try:
-            sheet = wb.sheet_by_name('metadata')
-            rows = list(sheet.get_rows())
+            sheet = wb['metadata']
+            rows = list(sheet.iter_rows())
             for row in rows:
                 if row[0].value == 'version':
                     version = row[1].value
@@ -851,55 +899,26 @@ class XLRDTableHandler(TableHandler):
         except:
             logger.info(f'could not find a sheet with name "metadata" in workbook. defaulting to v2')
 
-        for _sheet_name in _sheet_names:
-            if _sheet_name == 'metadata':
-                continue
-            sheet = wb.sheet_by_name(_sheet_name)
-            rows = list(sheet.get_rows())
-            header = [cell.value for cell in rows[0]]
+        table_visitor_partial = partial(self.table_visitor, wb=wb, sheet_names=_sheet_names,
+                                        definitions=definitions, id_flag=id_flag)
 
-            if header[0] != 'variable':
-                continue
+        table_visitor_partial(visitor_function=self.ref_date_handling)
+        if id_flag:
+            table_visitor_partial(visitor_function=self.add_ids)
+            wb.save(filename)
 
-            for i, row in enumerate(rows[1:]):
-                values = {}
-                for key, cell in zip(header, row):
-                    values[key] = cell.value
+        res = []
+        for var_set in definitions.values():
+            for scenario_var in var_set.values():
+                res.append(scenario_var)
 
-                if not values['variable']:
-                    # logger.debug(f'ignoring row {i}: {row}')
-                    continue
-
-                if 'ref date' in values and values['ref date']:
-                    if isinstance(values['ref date'], float):
-                        values['ref date'] = datetime.datetime(*xldate_as_tuple(values['ref date'], wb.datemode))
-                        if values['ref date'].day != 1:
-                            logger.warning(f'ref date truncated to first of month for variable {values["variable"]}')
-                            values['ref date'] = values['ref date'].replace(day=1)
-                    else:
-                        raise Exception(
-                            f"{values['ref date']} for variable {values['variable']} is not a date - "
-                            f"check spreadsheet value is a valid day of a month")
-                logger.debug(f'values for {values["variable"]}: {values}')
-                definitions.append(values)
-                scenario = values['scenario'] if values['scenario'] else "n/a"
-
-                if scenario in _definition_tracking[values['variable']]:
-
-                    logger.error(
-                        f"Duplicate entry for parameter "
-                        f"with name <{values['variable']}> and <{scenario}> scenario in sheet {_sheet_name}")
-                    raise ValueError(
-                        f"Duplicate entry for parameter "
-                        f"with name <{values['variable']}> and <{scenario}> scenario in sheet {_sheet_name}")
-
-                else:
-                    _definition_tracking[values['variable']][scenario] = 1
-        return definitions
+        return res
+        # return [definitions_ .values()]
+        # return definitions
 
 
 class XLWingsTableHandler(TableHandler):
-    def load_definitions(self, sheet_name, filename=None):
+    def load_definitions(self, sheet_name, filename=None, id_flag=False):
         import xlwings as xw
         definitions = []
         wb = xw.Book(fullname=filename)
@@ -914,7 +933,7 @@ class XLWingsTableHandler(TableHandler):
             if header[0] != 'variable':
                 continue
 
-            total_rows = XLRDTableHandler.get_sheet_range_bounds(filename, _sheet_name)
+            total_rows = OpenpyxlTableHandler.get_sheet_range_bounds(filename, _sheet_name)
             range = sheet.range((1, 1), (total_rows, len(header)))
             rows = range.rows
             for row in rows[1:]:
@@ -939,7 +958,7 @@ class TableParameterLoader(object):
 
        """
 
-    def __init__(self, filename, table_handler='xlrd', version=2, **kwargs):
+    def __init__(self, filename, table_handler='openpyxl', version=2, **kwargs):
         self.filename = filename
         self.definition_version = 2  # default - will be overwritten by handler
 
@@ -955,12 +974,9 @@ class TableParameterLoader(object):
             table_handler_instance = Xlsx2CsvHandler()
         if table_handler == 'xlwings':
             table_handler_instance = XLWingsTableHandler()
-        if table_handler == 'xlrd':
-            table_handler_instance = XLRDTableHandler()
-
         self.table_handler: TableHandler = table_handler_instance
 
-    def load_parameter_definitions(self, sheet_name: str = None):
+    def load_parameter_definitions(self, sheet_name: str = None, id_flag=False):
         """
         Load variable text from rows in excel file.
         If no spreadsheet arg is given, all spreadsheets are loaded.
@@ -983,22 +999,22 @@ class TableParameterLoader(object):
         :param sheet_name:
         :return: list of dicts with {header col name : cell value} pairs
         """
-        definitions = self.table_handler.load_definitions(sheet_name, filename=self.filename)
+        definitions = self.table_handler.load_definitions(sheet_name, filename=self.filename, id_flag=id_flag)
         self.definition_version = self.table_handler.version
         return definitions
 
-    def load_into_repo(self, repository: ParameterRepository = None, sheet_name: str = None):
+    def load_into_repo(self, repository: ParameterRepository = None, sheet_name: str = None, id_flag=False):
         """
         Create a Repo from an excel file.
         :param repository: the repository to load into
         :param sheet_name:
         :return:
         """
-        repository.add_all(self.load_parameters(sheet_name))
+        repository.add_all(self.load_parameters(sheet_name, id_flag=id_flag))
 
-    def load_parameters(self, sheet_name):
+    def load_parameters(self, sheet_name, id_flag=False):
 
-        parameter_definitions = self.load_parameter_definitions(sheet_name=sheet_name)
+        parameter_definitions = self.load_parameter_definitions(sheet_name=sheet_name, id_flag=id_flag)
         params = []
 
         param_name_map = param_name_maps[int(self.definition_version)]
